#ifndef KEYMAPPER_H
#define KEYMAPPER_H

#include <string>
#include <vector>

template<class T, class U>
class KeyMapper
{
public:
    KeyMapper ( const std::basic_string<T>& keyParameter, const std::vector<U>& toMapParameter );
    void map ( std::vector<U>& resultVector );
    void unMap ( std::vector<U>& resultVector );

private:
    std::vector<U> toMap;
    std::vector<size_t> mappingVector;
    std::vector<size_t> invMappingVector;

    void handleMap ( std::vector<U>& resultVector, const std::vector<size_t>& mappingVectorParameter );
    void getMapChunk ( std::basic_string<T>& keyParameter, std::vector<size_t>& mapChunk );
    void getInvMapChunk ( std::vector<size_t>& mapChunk, std::vector<size_t>& invMapChunk );
    void quicksortToSortToMapChunk ( std::vector<size_t>& toSortParameter, size_t lo, size_t hi, std::vector<size_t>& mapChunkParameter);

};

template<class T, class U>
KeyMapper<T, U>::KeyMapper ( const std::basic_string<T>& keyParameter, const std::vector<U>& toMapParameter )
{
    if ( !keyParameter.size() )
        throw;

    toMap = toMapParameter;
    std::basic_string<T> keyFull = keyParameter;

    size_t toMapSize = toMap.size();
    size_t keySize = keyFull.size();
    size_t mapChunkFullCount = toMapSize / keySize;
    size_t mapChunkFragmentSize = toMapSize % keySize;

    std::vector<size_t> mapChunkFull;
    std::vector<size_t> invMapChunkFull;
    if ( mapChunkFullCount )
    {
        getMapChunk( keyFull, mapChunkFull );
        getInvMapChunk( mapChunkFull, invMapChunkFull );
    }

    std::vector<size_t> mapChunkFragment;
    std::vector<size_t> invMapChunkFragment;
    if ( mapChunkFragmentSize )
    {
        std::basic_string<T> keyFragment = keyParameter.substr( 0, mapChunkFragmentSize );
        getMapChunk( keyFragment, mapChunkFragment );
        getInvMapChunk( mapChunkFragment, invMapChunkFragment );
    }

    size_t x;
    for (x=0; (toMapSize - x) > mapChunkFragmentSize; x++)
    {
        mappingVector.push_back ( mapChunkFull.at( x % keySize ) + x/keySize*keySize );
        invMappingVector.push_back( invMapChunkFull.at( x % keySize ) + x/keySize*keySize );
    }
    for (; (toMapSize - x) > 0; x++)
    {
        mappingVector.push_back ( mapChunkFragment.at( x % keySize ) + x/keySize*keySize );
        invMappingVector.push_back ( invMapChunkFragment.at( x % keySize ) + x/keySize*keySize );
    }

}

template<class T, class U>
void KeyMapper<T, U>::map ( std::vector<U>& resultVector )
{
    handleMap( resultVector, mappingVector);
    toMap = resultVector;
}

template<class T, class U>
void KeyMapper<T, U>::unMap ( std::vector<U>& resultVector )
{
    handleMap( resultVector, invMappingVector);
    toMap = resultVector;
}

template<class T, class U>
void KeyMapper<T, U>::handleMap ( std::vector<U>& resultVector, const std::vector<size_t>& mappingVectorParameter )
{
    resultVector.clear();
    size_t resultVectorSize = mappingVectorParameter.size();
    for (size_t i = 0; i < resultVectorSize; i++)
    {
        resultVector.push_back ( toMap.at( mappingVectorParameter.at( i ) ) );
    }
}

template<class T, class U>
void KeyMapper<T, U>::getMapChunk ( std::basic_string<T>& keyParameter, std::vector<size_t>& mapChunk )
{
    mapChunk.clear();
    size_t mapChunkSize = keyParameter.size();
    std::vector<size_t> temp;
    for (size_t i = 0; i < mapChunkSize; i++)
    {
        mapChunk.push_back( i );
        temp.push_back( (size_t) keyParameter.at( i ));
    }
    quicksortToSortToMapChunk (temp, 0, mapChunkSize - 1, mapChunk);
}

template<class T, class U>
void KeyMapper<T, U>::getInvMapChunk ( std::vector<size_t>& mapChunk, std::vector<size_t>& invMapChunk )
{
    std::vector<size_t> mapChunkCopy = mapChunk;
    invMapChunk.clear();
    size_t invMapChunkSize = mapChunk.size();
    std::vector<size_t> temp;
    for (size_t i = 0; i < invMapChunkSize; i++)
    {
        invMapChunk.push_back( i );
        temp.push_back( (size_t) mapChunkCopy.at( i ));
    }
    quicksortToSortToMapChunk (mapChunkCopy, 0, invMapChunkSize - 1, invMapChunk);
}

template<class T, class U>
void KeyMapper<T, U>::quicksortToSortToMapChunk ( std::vector<size_t>& toSortParameter, size_t lo, size_t hi, std::vector<size_t>& mapChunkParameter)
{
    size_t i = lo, j = hi, k, h;
    size_t x = toSortParameter.at ( ( lo + hi ) / 2 );

    do
    {
        while ( toSortParameter.at ( i ) < x )
            i++;
        while ( toSortParameter.at ( j ) > x )
        {

            if ( j )
                j--;
        }
        if ( i <= j )
        {
            h = toSortParameter.at( i );
            toSortParameter[i] = toSortParameter.at ( j );
            toSortParameter[j] = h;

            k = mapChunkParameter.at( i );
            mapChunkParameter[ i ]= mapChunkParameter.at( j );
            mapChunkParameter[ j ] = k;

            i++;
            if ( j )
                j--;
        }
    } while ( i <= j );

    if ( lo < j ) quicksortToSortToMapChunk(toSortParameter, lo, j, mapChunkParameter);
    if ( i < hi ) quicksortToSortToMapChunk(toSortParameter, i, hi, mapChunkParameter);
}

#endif
