#ifndef SLICECLIPTRANSFORMER_H
#define SLICECLIPTRANSFORMER_H

#include <cmath>

#include "dir.h"
#include "enum.h"
#include "samplemath.h"
#include "slicetransformer.h"
#include "slicenormalizetransformer.h"
#include "sliceffttransformer.h"
#include "wrapcsound.h"
#include "wraptempfilename.h"

template<class T>
class SliceClipTransformer : public SliceTransformer<T>
{
public:
    SliceClipTransformer (
        double cutoff,
        double distortRange,
        double amplificationFactor,
        int distortType = DISTORT_PROPORTIONAL,//DISTORT_70SFUZZGUITAR
        size_t feedback = 0,
        double toneShaping1 = 1.0,
        double toneShaping2 = 0.0,
        size_t channels = 2 );
    void transform ( std::vector<T> *leftChannel, std::vector<T> *rightChannel );
    void transformChannel ( std::vector<T> *channel );
private:
    double cutoff, distortRange, amplificationFactor;
    int distortType;
    size_t feedback;
    double toneShaping1, toneShaping2;
    size_t channels;
    T  distortSampleProportional( T sample, size_t pos );
};

template<class T>
SliceClipTransformer<T>::SliceClipTransformer (
    double cutoffParameter,
    double distortRangeParameter,
    double amplificationFactorParameter,
    int distortTypeParameter,
    size_t feedbackParameter,
    double toneShapingParameter1,
    double toneShapingParameter2,
    size_t channelsParameter ) :
        cutoff ( cutoffParameter ),
        distortRange ( distortRangeParameter ),
        amplificationFactor ( amplificationFactorParameter ),
        distortType ( distortTypeParameter ),
        feedback ( feedbackParameter ),
        toneShaping1 ( toneShapingParameter1 ),
        toneShaping2 ( toneShapingParameter2 ),
        channels ( channelsParameter )
{
    if ( distortRange > cutoff)
        throw;
    if ( amplificationFactor < 0 )
        throw;
    if ( ( amplificationFactor * cutoff ) > SHRT_MAX )
        throw;
}


template<class T>
void SliceClipTransformer<T>::transform (
    std::vector<T> *leftChannel,
    std::vector<T> *rightChannel )
{

    SliceNormalizeTransformer<T> snt ( 0.75 );
    snt.transform ( leftChannel, rightChannel );

    for ( size_t i = 0; i < feedback; i++ )
    {
        SliceNormalizeTransformer<T> snt1 ( 0.99 );
        snt.transform ( leftChannel, rightChannel );

        SliceFFTTransformer<T> lpf (
            FFT_LOWPASS,
            0.0,
            0.0,
            toneShaping1,
            toneShaping2
        );
        lpf.transform ( leftChannel, rightChannel );

        transformChannel ( leftChannel );
        if ( rightChannel )
            transformChannel ( rightChannel );

    }

}

template<class T>
void SliceClipTransformer<T>::transformChannel ( std::vector<T> *channel )
{
    size_t channelSize;
    channelSize = channel->size();

    for ( size_t i = 0; i < channelSize; i++ )
    {
        (*channel) [ i ] = ( distortSampleProportional ( channel->at ( i ), i ));
    }
}

template<class T>
T SliceClipTransformer<T>::distortSampleProportional ( T sample, size_t pos )
{
    double absSample, linearRange;
    absSample =  fabs ( sample );
    linearRange = cutoff - distortRange;
    if ( ! ( absSample > linearRange ) )
        return amplificationFactor * sample;
    double distortFactor, absSampleDistorted, sinFactor;
    distortFactor = distortRange / ( SHRT_MAX - linearRange );
    sinFactor = sin ( 2 * PI / 2205 * pos );
    absSampleDistorted = linearRange + distortFactor * sinFactor * ( absSample - linearRange );
    return amplificationFactor * SampleMath::sign ( sample ) * absSampleDistorted;
}

#endif
