#ifndef SLICECOMPRESSTRANSFORMER_H
#define SLICECOMPRESSTRANSFORMER_H

#include <string>
#include <cmath>
#include <iostream>

#include "fftreal/FFTReal.h"

#include "functor.h"
#include "samplemath.h"
#include "slicetransformer.h"
#include "slicegatetransformer.h"
#include "slicenormalizetransformer.h"

template<class T>
class SliceCompressTransformer : public SliceTransformer<T>
{
public:
    SliceCompressTransformer ( double cutoff, double compressRange, size_t channels = 2 );
    void transform ( std::vector<T> *leftChannel, std::vector<T> *rightChannel );
    void transformChannel ( std::vector<T>& channel );
private:
    double cutoffInit, compressRangeInit;
    T compressMagnitudeProportional ( T magnitude, T peakLevel, T cutoff, T compressRange );
    size_t channels;
};


template<class T>
SliceCompressTransformer<T>::SliceCompressTransformer
( double cutoffInitParameter, double compressRangeInitParameter, size_t channelsParameter ):
        cutoffInit (  cutoffInitParameter ), compressRangeInit ( compressRangeInitParameter ),
        channels ( channelsParameter )
{}

template<class T>
void SliceCompressTransformer<T>::transform ( std::vector<T> *leftChannel, std::vector<T> *rightChannel )
{
    transformChannel ( *leftChannel );

    if ( rightChannel )
        transformChannel ( *rightChannel  );

    SliceNormalizeTransformer<T> snt;
    snt.transform ( leftChannel, rightChannel );

    SliceGateTransformer<T> sgt (
        ( rightChannel ) ? 2 * leftChannel->size() : leftChannel->size(),
        64,
        channels );
    sgt.transform ( leftChannel, rightChannel );
}

template<class T>
void SliceCompressTransformer<T>::transformChannel ( std::vector<T>& channel )
{
    long channelSize, zeroPadding, segmentLen, segmentLenHalf;
    channelSize = (long) channel.size();
    segmentLen = SampleMath::getNextPower2( channelSize );
    segmentLenHalf = segmentLen / 2;
    zeroPadding = segmentLen - channelSize;

    std::vector<T> result;

    while ( zeroPadding )
    {
        channel.push_back ( 0.0 );
        zeroPadding--;
    }

    //fre = 0.025;//0.05; // values between 0.0(0Hz) and 0.5 (22050Hz)
    //dist = 0.002;//0.2;  // values between 0.0001(4.41) and 0.2 (8820)

    FFTReal<T> fft ( segmentLen );
    std::vector <T> cf (segmentLen);
    std::vector <T> cy (segmentLen);
    fft.do_fft (&cf [0], &channel [0]);

    //double dist, freqSave;
    //freqSave = cf [ segmentLenHalf ];

    double f, peakLevel;
    peakLevel = 0;

    for (int x = 0 ; x < segmentLenHalf ; x++)
        peakLevel = ( cf[x] > peakLevel ) ? cf[x] : peakLevel;

    double cutoff, compressRange;
    cutoff = cutoffInit * peakLevel;
    compressRange = compressRangeInit * peakLevel;


    for (int x = 0 ; x < segmentLenHalf ; x++)
    {
        f = (double) x / (double) segmentLen ;

        cf [x] = compressMagnitudeProportional ( cf [ x ], peakLevel, cutoff, compressRange );

        cf [segmentLenHalf + x] = compressMagnitudeProportional ( cf [ segmentLenHalf + x ], peakLevel, cutoff, compressRange );
    }

    //dist = (0.5 - bandHi) / slopeHi;
    //cf [segmentLenHalf] = freqSave * exp ( -dist * dist );

    fft.do_ifft (&cf [0], &cy [0]);
    fft.rescale (&cy [0]);
    result.insert( result.end(), cy.begin(), cy.end() );

    zeroPadding = segmentLen - channelSize;

    while ( zeroPadding )
    {
        result.pop_back ();
        zeroPadding--;
    }

    channel = result;

}

template<class T>
T SliceCompressTransformer<T>::compressMagnitudeProportional ( T magnitude, T peakLevel, T cutoff, T compressRange )
{
    double absMagnitude, linearRange;
    absMagnitude =  fabs ( magnitude );
    linearRange = cutoff - compressRange;
    if ( ! ( absMagnitude > linearRange ) )
        return magnitude;
    double compressFactor, absMagnitudeDistorted;
    compressFactor = compressRange / ( peakLevel - linearRange );
    absMagnitudeDistorted = linearRange + compressFactor * ( absMagnitude - linearRange );
    return SampleMath::sign ( magnitude ) * absMagnitudeDistorted;
}

#endif

