#ifndef SLICEDELAYPHASETRANSFORMER_H
#define SLICEDELAYPHASETRANSFORMER_H

#include <string>
#include <cmath>
#include <iostream>

#include "enum.h"
#include "functor.h"
#include "samplemath.h"
#include "slicetransformer.h"
#include "slicegatetransformer.h"


template<class T>
class SliceDelayPhaseTransformer : public SliceTransformer<T>
{
public:
    SliceDelayPhaseTransformer (
        double phaseDistanceOffset = 440.0 / 22050.0,
        double phaseDistanceRange = 1600.0 / 22050.0,
        int lfoType = FFT_LFO_SINPLUS,
        double lfoDepth = 0.5,
        double lfoPeriodsPerSlice = 0.1,
        size_t channels = 2,
        double lfoPhaseOffset = 0,
        double gain = 0.7
    );
    void transform ( std::vector<T> *leftChannel, std::vector<T> *rightChannel );
    void transformChannel ( std::vector<T> *channel );//std::vector<T>& channel );
    void updateDelayCoefficient ( T delay, T *delayCoefficient );
    T filter ( T input, T *delayCoefficient, T *outputMemory );
private:
    double phaseDistanceOffset, phaseDistanceRange;
    int lfoType;
    double ( *lfo ) ( double f );
    double lfoDepth, lfoPeriodsPerSlice;
    size_t channels;
    double lfoPhaseOffset, gain;
};

template<class T>
SliceDelayPhaseTransformer<T>::SliceDelayPhaseTransformer
(
    double phaseDistanceOffsetParameter,
    double phaseDistanceRangeParameter,
    int lfoTypeParameter,
    double lfoDepthParameter,
    double lfoPeriodsPerSliceParameter,
    size_t channelsParameter,
    double gainParameter,
    double  lfoPhaseOffsetParameter ) :
        phaseDistanceOffset ( phaseDistanceOffsetParameter ),
        phaseDistanceRange ( phaseDistanceRangeParameter ),
        lfoType ( lfoTypeParameter ),
        lfoDepth ( lfoDepthParameter ),
        lfoPeriodsPerSlice ( lfoPeriodsPerSliceParameter ),
        channels ( channelsParameter ),
        lfoPhaseOffset ( lfoPhaseOffsetParameter ),
        gain ( gainParameter )
{
    switch ( lfoType )
    {
    case FFT_LFO_SINPOW2:
        lfo = sinPow2;
        break;
    case FFT_LFO_SINPLUS:
        lfo = sinPlus;
        break;
    case FFT_LFO_SAWTOOTH:
        lfo = aaSawTooth;
        break;
    case FFT_LFO_SQUARE:
        lfo = aaSquare;
        break;
    case FFT_LFO_TRIANGLE:
        lfo = aaTriangle;
        break;
    case FFT_LFO_SAMPLEANDHOLD:
        lfo = sampleAndHold;
        break;
    case FFT_LFO_NONE:
        ;
    default:
        lfo = passThrough;
        break;
    }

}

template<class T>
void SliceDelayPhaseTransformer<T>::transform ( std::vector<T> *leftChannel, std::vector<T> *rightChannel )
{
    transformChannel ( leftChannel );
    if ( rightChannel )
        transformChannel ( rightChannel );

    SliceGateTransformer<T> sgt (
        ( rightChannel ) ?
        leftChannel->size() + rightChannel->size() :
        leftChannel->size(),
        64,
        channels );
    sgt.transform ( leftChannel, rightChannel );
}

template<class T>
void SliceDelayPhaseTransformer<T>::updateDelayCoefficient ( T delay, T *delayCoefficient )
{
    *delayCoefficient = ( 1.0 - delay ) / ( 1.0 + delay  );
}

template<class T>
T SliceDelayPhaseTransformer<T>::filter( T input, T *delayCoefficient, T *outputMemory )
{
    T output = *outputMemory - ( *delayCoefficient * input );
    *outputMemory = input + ( *delayCoefficient * output );
    return output;
}

template<class T>
void SliceDelayPhaseTransformer<T>::transformChannel ( std::vector<T> *channel )
{
    size_t channelSize = channel->size();

    std::vector<std::pair<T, T> > allpassFilterVector; //std::pair of delayCoefficient, outputMemory

    for ( size_t i = 0; i < 6; i++ )
        allpassFilterVector.push_back ( std::make_pair ( 0.0, 0.0 ) );

    T delay, currentSample, filterOutput, filterOutputMemory;
    filterOutputMemory = 0.0;
    double lfoFrequency = lfoPeriodsPerSlice / ( (double) channelSize );
    size_t allpassFilterVectorSize = allpassFilterVector.size();
    for ( size_t i = 0; i < channelSize; i++ )
    {
        delay = phaseDistanceOffset + phaseDistanceRange * lfo ( 2 * PI * lfoFrequency * (double) i );

        for ( size_t j = 0; j < allpassFilterVectorSize; j++ )
            updateDelayCoefficient ( delay, &allpassFilterVector.at ( j ).first );

        currentSample = channel->at ( i );

        filterOutput = filter (
                           filter (
                               filter (
                                   filter (
                                       filter (
                                           filter ( currentSample + gain * filterOutputMemory,
                                                    &allpassFilterVector[ 0 ].first,
                                                    &allpassFilterVector[ 0 ].second ),
                                           &allpassFilterVector[ 1 ].first,
                                           &allpassFilterVector[ 1 ].second ),
                                       &allpassFilterVector[ 2 ].first,
                                       &allpassFilterVector[ 2 ].second ),
                                   &allpassFilterVector[ 3 ].first,
                                   &allpassFilterVector[ 3 ].second ),
                               &allpassFilterVector[ 4 ].first,
                               &allpassFilterVector[ 4 ].second ),
                           &allpassFilterVector[ 5 ].first,
                           &allpassFilterVector[ 5 ].second );

        filterOutputMemory = filterOutput;
        (*channel)[ i ] = currentSample + lfoDepth * filterOutput;
    }
}

#endif
