#ifndef SLICEDELAYTRANSFORMER_H
#define SLICEDELAYTRANSFORMER_H

#include <string>
#include <cmath>
#include <iostream>

#include "enum.h"
#include "functor.h"
#include "samplemath.h"
#include "slicetransformer.h"
#include "slicenormalizetransformer.h"
#include "slicegatetransformer.h"


template<class T>
class SliceDelayTransformer : public SliceTransformer<T>
{
public:
    SliceDelayTransformer (
        double delay = 0.25,
        size_t stepNo = 1,
        double gain = 0.5,
        size_t feedbackLoopNo = 1,
        size_t channels = 2 );
    void transform ( std::vector<T> *leftChannel, std::vector<T> *rightChannel );
    void transformChannel ( std::vector<T> *channel );
private:
    double delay;
    size_t stepNo;
    double gain;
    size_t feedbackLoopNo, channels;
    void ( *addDelay ) ( T *t, size_t len, size_t distance, double gain );
    static void addDelayOneTap ( T *t, size_t len, size_t distance, double gain );
    static void addDelayTwoTap ( T *t, size_t len, size_t distance, double gain );
    static void addDelayThreeTap ( T *t, size_t len, size_t distance, double gain );
    static void addDelayFourTap ( T *t, size_t len, size_t distance, double gain );
};

template<class T>
SliceDelayTransformer<T>::SliceDelayTransformer (
    double delayParameter,
    size_t stepNoParameter,
    double gainParameter,
    size_t feedbackLoopNoParameter,
    size_t channelsParameter ):
        delay( delayParameter ),
        stepNo ( stepNoParameter ),
        gain ( gainParameter ),
        feedbackLoopNo ( feedbackLoopNoParameter ),
        channels ( channelsParameter )
{
    if ( !stepNo || delay <= 0.0 || delay > 1.0 )
    {
        std::cout << "invalid parameter" << std::endl;
        return;
    }

    size_t tmp;
    tmp = ( size_t ) ( 1.0 / delay );

    stepNo = ( tmp < stepNo ) ? tmp : stepNo;

    switch ( stepNo )
    {
    case 1:
        addDelay = addDelayOneTap;
        break;
    case 2:
        addDelay = addDelayTwoTap;
        break;
    case 3:
        addDelay = addDelayThreeTap;
        break;
    case 4:
        addDelay = addDelayFourTap;
        break;
    default:
        addDelay = addDelayFourTap;
        break;
    }
}

template<class T>
void SliceDelayTransformer<T>::transform ( std::vector<T> *leftChannel, std::vector<T> *rightChannel )
{
    transformChannel ( leftChannel );
    if ( rightChannel )
        transformChannel ( rightChannel );

    double peakLevel = SampleMath::getPeakLevel ( leftChannel, rightChannel );
    if ( peakLevel > DOUBLE_SHRT_MAX )
    {
        SliceNormalizeTransformer<T> snt;
        snt.transform ( leftChannel, rightChannel );
    }

    SliceGateTransformer<T> sgt (
        ( rightChannel ) ? 2 * leftChannel->size() : leftChannel->size(),
        64,
        channels );
    sgt.transform ( leftChannel, rightChannel );
}

template<class T>
void SliceDelayTransformer<T>::transformChannel ( std::vector<T> *channel )
{
    size_t channelSize = channel->size();
    size_t delayDistance;
    double currentGain = gain;

    delayDistance = ( size_t ) ( delay * channelSize );
    for ( size_t i = 0; i <= feedbackLoopNo; i++ )
    {
        currentGain *= 0.9;
        addDelay ( &((*channel)[0]), channelSize, delayDistance, currentGain );
    }
}

template<class T>
void SliceDelayTransformer<T>::addDelayOneTap ( T *t, size_t len, size_t distance, double gain )
{
    T *guard, *it;
    guard = t + len;

    it = t + distance;

    for ( ; it != guard; it++ )
    {
        *it += gain * ( *( it - distance ) );
    }

}

template<class T>
void SliceDelayTransformer<T>::addDelayTwoTap ( T *t, size_t len, size_t distance, double gain )
{
    T *guard, *it;

    guard = t + 2 * distance;
    for ( it = t + distance; it != guard; it++ )
    {
        *it += gain * ( *( it - distance ) );
    }

    guard = t + len;
    for ( it = t + 2 * distance; it != guard; it++ )
    {
        *it += gain * ( *( it - distance ) ) + gain * ( *( it - 2 * distance ) );
    }
}

template<class T>
void SliceDelayTransformer<T>::addDelayThreeTap ( T *t, size_t len, size_t distance, double gain )
{
    T *guard, *it;

    guard = t + 2 * distance;
    for ( it = t + distance; it != guard; it++ )
    {
        *it += gain * ( *( it - distance ) );
    }

    guard = t + 3 * distance;
    for ( it = t + 2 * distance; it != guard; it++ )
    {
        *it += gain * ( *( it - distance ) ) + gain * ( *( it - 2 * distance ) );
    }

    guard = t + len;
    for ( it = t + 3 * distance; it != guard; it++ )
    {
        *it += gain * ( *( it - distance ) ) + gain * ( *( it - 2 * distance ) ) + gain * ( *( it - 3 * distance ) );
    }
}

template<class T>
void SliceDelayTransformer<T>::addDelayFourTap ( T *t, size_t len, size_t distance, double gain )
{
    T *guard, *it;

    guard = t + 2 * distance;
    for ( it = t + distance; it != guard; it++ )
    {
        *it += gain * ( *( it - distance ) );
    }

    guard = t + 3 * distance;
    for ( it = t + 2 * distance; it != guard; it++ )
    {
        *it += gain * ( *( it - distance ) ) + gain * ( *( it - 2 * distance ) );
    }

    guard = t + 4 * distance;
    for ( it = t + 3 * distance; it != guard; it++ )
    {
        *it += gain * ( *( it - distance ) ) + gain * ( *( it - 2 * distance ) ) + gain * ( *( it - 3 * distance ) );
    }

    guard = t + len;
    for ( it = t + 4 * distance; it != guard; it++ )
    {
        *it += gain * ( *( it - distance ) ) + gain * ( *( it - 2 * distance ) ) + gain * ( *( it - 3 * distance ) ) + gain * ( *( it - 4 * distance ) );
    }
}

#endif
