#ifndef SLICEDELAYVARIABLETRANSFORMER_H
#define SLICEDELAYVARIABLETRANSFORMER_H

#include <string>
#include <set>
#include <cmath>
#include <iostream>
#include <stdexcept>

#include "enum.h"
#include "functor.h"
#include "samplemath.h"
#include "slicetransformer.h"
#include "slicenormalizetransformer.h"
#include "slicegatetransformer.h"

enum delayLfoTypes { DELAY_LFO_SINPOW2, DELAY_LFO_TRIANGLE };

template<class T>
class SliceDelayVariableTransformer : public SliceTransformer<T>
{
public:
    SliceDelayVariableTransformer (
        size_t delay = 16,
        size_t feedbackLoopNo = 1,
        size_t channels = 2,
        int lfoType = DELAY_LFO_SINPOW2,
        double gain = 0.5,
        double lfoDepth = 1,
        double lfoPeriodsPerSlice = 0.125 );
    void transform ( std::vector<T> *leftChannel, std::vector<T> *rightChannel );
    void transformChannel ( std::vector<T> *channel );
private:
    size_t delay, feedbackLoopNo, channels;
    int lfoType;
    double gain, lfoDepth, lfoPeriodsPerSlice;
    double ( *lfo ) ( double frequency );
    std::vector<size_t> delayOffsetVector;
    std::set<int> overlapSet;
};

template<class T>
SliceDelayVariableTransformer<T>::SliceDelayVariableTransformer (
    size_t delayParameter,
    size_t feedbackLoopNoParameter,
    size_t channelsParameter,
    int lfoTypeParameter,
    double gainParameter,
    double lfoDepthParameter,
    double lfoPeriodsPerSliceParameter ) :
        delay ( delayParameter ),
        feedbackLoopNo ( feedbackLoopNoParameter ),
        channels ( channelsParameter ),
        lfoType ( lfoTypeParameter ),
        gain ( gainParameter ),
        lfoDepth ( lfoDepthParameter ),
        lfoPeriodsPerSlice ( lfoPeriodsPerSliceParameter )
{
    switch ( lfoType )
    {
    case DELAY_LFO_SINPOW2:
        lfo = sinPow2;
        break;
    case DELAY_LFO_TRIANGLE:
        lfo = aaTriangle;
        break;
    default:
        lfo = sinPow2;
    }
}

template<class T>
void SliceDelayVariableTransformer<T>::transform ( std::vector<T> *leftChannel, std::vector<T> *rightChannel )
{
    transformChannel ( leftChannel );

    if ( rightChannel )
        transformChannel ( rightChannel );


    double peakLevel = SampleMath::getPeakLevel ( leftChannel, rightChannel );
    if ( peakLevel > DOUBLE_SHRT_MAX )
    {
        SliceNormalizeTransformer<T> snt;
        snt.transform ( leftChannel, rightChannel );
    }

    SliceGateTransformer<T> sgt (
        ( rightChannel ) ? 2 * leftChannel->size() : leftChannel->size(),
        64,
        channels );
    sgt.transform ( leftChannel, rightChannel );
}


template<class T>
void SliceDelayVariableTransformer<T>::transformChannel ( std::vector<T> *channel )
{
    size_t channelSize = channel->size();
    size_t delayOffset;
    double currentGain, lfoFrequency, lfoFactor;
    currentGain = gain;
    lfoFrequency = ( double ) lfoPeriodsPerSlice / (double) channelSize;

    delayOffsetVector.clear();
    overlapSet.clear();

    size_t delayOffsetMemory = ( size_t ) ( ( 1.0 - lfoDepth ) +
                                            lfoDepth * lfo ( ( 2 * PI * lfoFrequency * (double) 0 ) + ( 0.25 * PI ) ) * delay );
    for ( size_t j = 0; j < channelSize; j++ )
    {
        lfoFactor = ( 1.0 - lfoDepth ) + lfoDepth * lfo ( ( 2 * PI * lfoFrequency * (double) j ) + ( 0.25 * PI ) );
        delayOffset = ( size_t ) ( lfoFactor * delay );

        delayOffsetVector.push_back ( delayOffset );
        if ( delayOffset != delayOffsetMemory )
        {
            overlapSet.insert ( j - 1 );
            overlapSet.insert ( -j );
        }
        delayOffsetMemory = delayOffset;

    }

    for ( size_t i = 0; i <= feedbackLoopNo; i++ )
    {
        currentGain *= 0.9;
        for ( size_t j =0; j < channelSize; j++ )
        {
            if ( ( j > delayOffsetVector.at( j ) ) && ( j < channelSize - 1 ) && ( j > 0 ) )
            {
                if ( !overlapSet.count ( j ) && !overlapSet.count ( -j ) )
                    (*channel) [ j ] += currentGain * channel->at ( j - delayOffsetVector.at( j ) );
                else if ( overlapSet.count ( -j ) )
                    (*channel) [ j ] += currentGain * ( 0.5 * ( channel->at ( j - delayOffsetVector.at( j ) ) + channel->at ( j - delayOffsetVector.at( j - 1 ) ) ) );
                else if ( overlapSet.count ( j ) )
                    (*channel) [ j ] += currentGain * 0.5 * ( channel->at ( j - delayOffsetVector.at( j ) ) + channel->at ( j - delayOffsetVector.at( j + 1 ) ) );
            }
        }
    }

}

#endif
