#ifndef SLICEFFTTRANSFORMER_H
#define SLICEFFTTRANSFORMER_H

#include <string>
#include <cmath>
#include <iostream>

#include "fftreal/FFTReal.h"

#include "enum.h"
#include "functor.h"
#include "samplemath.h"
#include "slicetransformer.h"
#include "slicegatetransformer.h"
#include "sliceffttransformer.h"


template<class T>
class SliceFFTTransformer : public SliceTransformer<T>
{
public:
    SliceFFTTransformer ( int filterType,
                          double bandLo, double slopeLo,
                          double bandHi, double slopeHi,
                          size_t channels = 2 );
    void transform ( std::vector<T> *leftChannel, std::vector<T> *rightChannel );
    void transformChannel ( std::vector<T>& channel );
protected:
    int filterType;
    double ( *filter ) ( double frequency,
                         double bandLo, double slopeLo,
                         double bandHi, double slopeHi);
private:
    double bandLo, slopeLo, bandHi, slopeHi;
    size_t channels;
};


template<class T>
SliceFFTTransformer<T>::SliceFFTTransformer
( int filterTypeParameter,
  double bandLoParameter, double slopeLoParameter,
  double bandHiParameter, double slopeHiParameter,
  size_t channelsParameter ) :
        filterType (  filterTypeParameter ), bandLo ( bandLoParameter ), slopeLo ( slopeLoParameter ),
        bandHi ( bandHiParameter ), slopeHi ( slopeHiParameter ), channels ( channelsParameter )
{
    switch ( filterType )
    {
    case ( FFT_LOWPASS ):
        filter = fftLowpassFilter;
        break;
    case ( FFT_HIGHPASS ):
        filter = fftHighpassFilter;
        break;
    case ( FFT_BANDPASS ):
        filter = fftBandpassFilter;
        break;
    case ( FFT_BANDSTOP ):
        filter = fftBandstopFilter;
        break;
    default:
        filter = fftLowpassFilter;
        break;
    }
}

template<class T>
void SliceFFTTransformer<T>::transform ( std::vector<T> *leftChannel, std::vector<T> *rightChannel )
{
    if ( channels == 1 )//tbd: inspect mono
        return;
    transformChannel ( *leftChannel );
    if ( rightChannel )
        transformChannel ( *rightChannel );

    SliceGateTransformer<T> sgt (
        ( rightChannel ) ? leftChannel->size() + rightChannel->size() : leftChannel->size(),
        64,
        channels );
    sgt.transform ( leftChannel, rightChannel );
}

template<class T>
void SliceFFTTransformer<T>::transformChannel ( std::vector<T>& channel )
{
    std::vector<T> result;

    long channelSize, segmentLen, segmentLenHalf, zeroPadding, zeroPaddingRemove;
    channelSize = (long) channel.size();
    segmentLen = SampleMath::getNextPower2( channelSize );
    segmentLenHalf = segmentLen / 2;
    zeroPadding = segmentLen - channelSize;
    zeroPaddingRemove = zeroPadding;

    while ( zeroPadding )
    {
        channel.push_back ( 0.0 );
        zeroPadding--;
    }

    double f, dist, wt, freqSave;

    //fre = 0.025;//0.05; // values between 0.0(0Hz) and 0.5 (22050Hz)
    //dist = 0.002;//0.2;  // values between 0.0001(4.41) and 0.2 (8820)

    FFTReal<T> fft ( segmentLen );
    std::vector <T> cf (segmentLen);
    std::vector <T> cy (segmentLen);
    fft.do_fft (&cf [0], &channel [0]);

    freqSave = cf.at (segmentLenHalf);

    for (int x = 0 ; x < segmentLenHalf ; x++)
    {
        f = (double) x / (double) segmentLen ;

        wt = filter ( f, bandLo, slopeLo, bandHi, slopeHi );

        cf [x] = cf.at(x) * wt;
        cf [segmentLenHalf + x] = cf.at(segmentLenHalf + x) * wt;
    }

    dist = (0.5 - bandHi) / slopeHi;
    cf [segmentLenHalf] = freqSave * exp ( -dist * dist );

    fft.do_ifft (&cf [0], &cy [0]);
    fft.rescale (&cy [0]);
    result.insert( result.end(), cy.begin(), cy.end() );

    while ( zeroPaddingRemove )
    {
        result.pop_back ();
        zeroPaddingRemove--;
    }

    channel = result;
}

#endif
