#ifndef SLICEFFTWINDOWEDTRANSFORMER_H
#define SLICEFFTWINDOWEDTRANSFORMER_H

#include <string>
#include <cmath>
#include <iostream>
#include "fftreal/FFTReal.h"

#include "enum.h"
#include "functor.h"
#include "samplemath.h"
#include "slicetransformer.h"
#include "slicecliptransformer.h"
#include "slicegatetransformer.h"
#include "slicenormalizetransformer.h"

template<class T>
class SliceFFTWindowedTransformer : public SliceTransformer<T>
{
public:
    SliceFFTWindowedTransformer (
        int filterType,
        int filterModulationType,
        int filterLfoType,
        double bandLo, double slopeLo,
        double bandHi, double slopeHi, long windowSize = 1024,
        double modulationDepth = 0.0,
        double lfoDepth = 0.0,
        double resonanceMax = 0.0,
        size_t lfoPeriodsPerSlice = 0,
        size_t channels = 2 );
    void transform ( std::vector<T> *leftChannel, std::vector<T> *rightChannel );
    void transformChannel ( std::vector<T>& channel );

private:
    int filterType, filterModulationType, filterLfoType;
    double ( *filter ) ( double frequency,
                         double bandLo, double slopeLo,
                         double bandHi, double slopeHi);
    double ( *modulation ) ( double f, double limit );
    double ( *lfo ) ( double f );
    double bandLo, slopeLo, bandHi, slopeHi;
    long windowSize;
    double modulationDepth, lfoDepth, resonanceMax;
    size_t lfoPeriodsPerSlice, channels;
};


template<class T>
SliceFFTWindowedTransformer<T>::SliceFFTWindowedTransformer
( int filterTypeParameter,
  int filterModulationTypeParameter,
  int filterLfoTypeParameter,
  double bandLoParameter, double slopeLoParameter,
  double bandHiParameter, double slopeHiParameter,
  long windowSizeParameter,
  double modulationDepthParameter,
  double lfoDepthParameter,
  double resonanceMaxParameter,
  size_t lfoPeriodsPerSliceParameter,
  size_t channelsParameter ) :
        filterType (  filterTypeParameter ),
        filterModulationType ( filterModulationTypeParameter ),
        filterLfoType ( filterLfoTypeParameter ),
        bandLo ( bandLoParameter ), slopeLo ( slopeLoParameter ),
        bandHi ( bandHiParameter ), slopeHi ( slopeHiParameter ),
        windowSize ( windowSizeParameter ),
        modulationDepth ( modulationDepthParameter ),
        lfoDepth ( lfoDepthParameter ),
        resonanceMax ( resonanceMaxParameter ),
        lfoPeriodsPerSlice ( lfoPeriodsPerSliceParameter ),
        channels ( channelsParameter )
{
    switch ( filterType )
    {
    case ( FFT_LOWPASS ):
        filter = fftLowpassFilter;
        break;
    case ( FFT_HIGHPASS ):
        filter = fftHighpassFilter;
        break;
    case ( FFT_BANDPASS ):
        filter = fftBandpassFilter;
        break;
    case ( FFT_BANDSTOP ):
        filter = fftBandstopFilter;
        break;
    default:
        filter = fftLowpassFilter;
        break;
    }

    switch ( filterModulationType )
    {
    case ( FFT_MOD_SWEEP_IN ):
        modulation = sweepIn;
        break;
    case ( FFT_MOD_SWEEP_OUT ):
        modulation = sweepOut;
        break;
    case ( FFT_MOD_NONE ):
        ;
    default:
        modulation = passThrough;
        break;
    }

    switch ( filterLfoType )
    {
    case FFT_LFO_SINPOW2:
        lfo = sinPow2;
        break;
    case FFT_LFO_SINPLUS:
        lfo = sinPlus;
        break;
    case FFT_LFO_SAWTOOTH:
        lfo = aaSawTooth;
        break;
    case FFT_LFO_SQUARE:
        lfo = aaSquare;
        break;
    case FFT_LFO_TRIANGLE:
        lfo = aaTriangle;
        break;
    case FFT_LFO_SAMPLEANDHOLD:
        lfo = sampleAndHold;
        break;
    case FFT_LFO_NONE:
        ;
    default:
        lfo = passThrough;
        break;
    }

}

template<class T>
void SliceFFTWindowedTransformer<T>::transform ( std::vector<T> *leftChannel, std::vector<T> *rightChannel )
{
    SliceNormalizeTransformer<T> snt ( 0.9 );
    snt.transform ( leftChannel, rightChannel );

    transformChannel ( *leftChannel );
    if ( rightChannel )
        transformChannel ( *rightChannel );

    SliceClipTransformer<T> sct ( ( double ) SHRT_MAX - 1000, 1000, 1 );
    sct.transform ( leftChannel, rightChannel );

    SliceGateTransformer<T> sgt (
        ( rightChannel ) ?
        leftChannel->size() + rightChannel->size() :
        leftChannel->size(),
        64,
        channels );
    sgt.transform ( leftChannel, rightChannel );
}

template<class T>
void SliceFFTWindowedTransformer<T>::transformChannel ( std::vector<T>& channel )
{
    std::vector<T> result;

    typename std::vector<T>::iterator resultIt;

    long channelSize, windowSizeHalf, crossFadeSize, windowOffset, windowNo, zeroPadding, zeroPaddingRemove;
    channelSize = channel.size();
    windowSizeHalf = windowSize / 2;
    crossFadeSize = windowSize / 2;
    windowOffset = windowSize - crossFadeSize;
    windowNo =  ( channelSize - windowSize ) / windowOffset + 1;
    zeroPadding = windowSize - ( channelSize - windowSize - ( windowNo * windowOffset ) );

    zeroPaddingRemove = zeroPadding;
    while ( zeroPadding )
    {
        channel.push_back ( 0.0 );
        zeroPadding--;
    }

    for ( size_t j = 0; j < channel.size(); j++ )
        result.push_back ( 0.0 );

    double f, dist, wt, freqSave;
    double lfoFrequency, lfoFactor, modulationFactor, resonanceFactor;

    lfoFrequency = lfoPeriodsPerSlice / ( (double) windowNo );

    int resonanceCount;
    bool resonanceDone;
    for ( long i = 0; i < windowNo; i++ )
    {
        FFTReal<T> fft ( windowSize );
        std::vector <T> cf (windowSize);
        std::vector <T> cy (windowSize);
        fft.do_fft (&cf [0], &channel [ i * windowOffset ]);

        freqSave = cf.at (windowSizeHalf);

        lfoFactor = ( 1.0 - lfoDepth ) + lfoDepth * lfo ( 2 * PI * lfoFrequency * (double) i );
        modulationFactor = ( 1.0 - modulationDepth ) + modulationDepth * modulation ( (double) i, ( double ) ( windowNo - 1 ) );

        resonanceCount = 0;
        resonanceDone = false;

        for (int x = 0 ; x < windowSizeHalf ; x++)
        {
            f = (double) x / (double) windowSize ;

            wt = filter (
                     f,
                     lfoFactor * modulationFactor * bandLo,
                     lfoFactor * modulationFactor * slopeLo,
                     lfoFactor * modulationFactor * bandHi,
                     lfoFactor * modulationFactor * slopeHi );

            if ( resonanceMax != 0.0 &&
                    !resonanceDone &&
                    ( (( wt < 1.0 ) && ( filterType == FFT_LOWPASS )) ||
                      (( wt > 0.99 ) && ( filterType == FFT_HIGHPASS )) ||
                      (( wt < 1.00 ) && ( filterType == FFT_BANDSTOP )) ||
                      (( wt > 0.99 ) && ( filterType == FFT_BANDPASS )) ) )
            {

                for ( int pos = -2; ( pos < 3 ) && ( x > 2 ) && ( x < windowSizeHalf - 2 ) ; pos++ )
                {
                    resonanceFactor = resonanceMax * 1.0 / ( 1.0 + pow ( pos, 2 ));
                    cf [x + pos] = cf.at( x + pos) * ( 1.0 + resonanceFactor );
                    cf [windowSizeHalf + x + pos] = cf.at(windowSizeHalf + x + pos) * ( 1.0 + resonanceFactor );
                }
                resonanceDone = true;

            }

            cf [x] = cf.at(x) * wt;
            cf [windowSizeHalf + x] = cf.at(windowSizeHalf + x) * wt;
        }

        dist = (0.5 - bandHi) / slopeHi;
        cf [windowSizeHalf] = freqSave * exp ( -dist * dist );

        fft.do_ifft (&cf [0], &cy [0]);
        fft.rescale (&cy [0]);


        for ( long j = 0; j < windowSize; j++ )
        {
            if ( i && ( j < crossFadeSize ) )
            {
                cy [ j ] = lConcaveSoftClipper ( j, 0, crossFadeSize ) * cy.at ( j );
            }
            else if ( ( i < ( windowNo - 1 ) ) && ( j > windowSize - crossFadeSize - 1 ) )
            {
                cy [ j ] = rConvexSoftClipper ( j, windowSize - crossFadeSize, windowSize ) * cy.at ( j );
            }
            result[ ( i * windowOffset ) + j ] = result.at( ( i * windowOffset ) + j ) + cy.at ( j );
        }
    }

    while ( zeroPaddingRemove )
    {
        result.pop_back ();
        zeroPaddingRemove--;
    }

    channel = result;

}

#endif
