#ifndef SLICETREMOLOTRANSFORMER_H
#define SLICETREMOLOTRANSFORMER_H

#include <string>
#include <cmath>
#include <iostream>

#include "enum.h"
#include "functor.h"
#include "samplemath.h"
#include "slicetransformer.h"


template<class T>
class SliceTremoloTransformer : public SliceTransformer<T>
{
public:
    SliceTremoloTransformer (
        int tremoloType = TREMOLO_SAWTOOTH,
        double tremoloDepth = 1.0,
        size_t tremoloPeriodsPerSlice = 2,
        int stereoMode = TREMOLO_SYNCHRONOUS,
        size_t channels = 2 );
    void transform ( std::vector<T> *leftChannel, std::vector<T> *rightChannel );
    void transformChannel ( std::vector<T> *channel, bool withOffset );
private:
    int tremoloType;
    double tremoloDepth;
    size_t tremoloPeriodsPerSlice;
    int stereoMode;
    double ( *lfo ) ( double frequency );
    double offset;
    size_t channels;
};

template<class T>
SliceTremoloTransformer<T>::SliceTremoloTransformer (
    int tremoloTypeParameter,
    double tremoloDepthParameter,
    size_t tremoloPeriodsPerSliceParameter,
    int stereoModeParameter,
    size_t channelsParameter ):
        tremoloType ( tremoloTypeParameter ),
        tremoloDepth ( tremoloDepthParameter ),
        tremoloPeriodsPerSlice ( tremoloPeriodsPerSliceParameter ),
        stereoMode ( stereoModeParameter ),
        channels ( channelsParameter )
{
    if ( tremoloDepth > 1.0 ||  tremoloDepth < 0 )
    {
        std::cout << "tremoloDepth out of bound - not in [ 0; 1 ]" << std::endl;
        return;
    }

    switch ( tremoloType )
    {
    case TREMOLO_SAWTOOTH:
        lfo = aaSawTooth;
        break;
    case TREMOLO_SQUARE:
        lfo = aaSquare;
        break;
    case TREMOLO_TRIANGLE:
        lfo = aaTriangle;
        break;
    case TREMOLO_SIN:
        lfo = sinPlus;
        break;
    case TREMOLO_SIN_POW2:
    default:
        lfo = sinPow2;
    }

    switch ( stereoMode )
    {
    case TREMOLO_SYNCHRONOUS:
        offset =  0;
        break;
    case TREMOLO_PING_PONG:
        offset = 0.5;
        break;
    case TREMOLO_LEFT_DELAYED:
        offset = -0.1;
        break;
    case TREMOLO_RIGHT_DELAYED:
        offset = 0.1;
        break;
    default:
        offset = 0;
    }

}

template<class T>
void SliceTremoloTransformer<T>::transform ( std::vector<T> *leftChannel, std::vector<T> *rightChannel )
{
    transformChannel ( leftChannel, true );
    if ( rightChannel )
        transformChannel ( rightChannel, false );
}


template<class T>
void SliceTremoloTransformer<T>::transformChannel ( std::vector<T> *channel, bool withOffset )
{
    size_t channelSize = channel->size();
    double tremoloFrequency, tremoloFactor, offsetTmp;

    tremoloFrequency = tremoloPeriodsPerSlice / ( (double) channelSize );
    offsetTmp = ( withOffset ) ? offset : 0.0;

    for ( size_t i = 0; i < channelSize; i++ )
    {
        tremoloFactor = ( 1.0 - tremoloDepth ) + tremoloDepth * lfo ( 2 * PI * ( ( tremoloFrequency * (double) i )  + offsetTmp ) );
        (*channel) [ i ] = tremoloFactor * channel->at ( i );
    }
}

#endif
