#include <iostream>
#include <memory>
#include <stdexcept>
#include <cstring>
#include <cmath>
#include <vector>
#include "sndfile.hh"
#include "wav16.h"
#include "samplemath.h"

Wav16::Wav16()
{
    channels = samplerate = 0;
    format = SF_FORMAT_WAV | SF_FORMAT_PCM_16;
    setState ( WAV16_INIT );
}

Wav16::~Wav16()
{
}

bool Wav16::read ( char *fname )
{
    clear();

    std::auto_ptr<SndfileHandle> sfh ( new SndfileHandle ( (const char*) fname ) );

    if ( !sfh.get() )
    {
        std::cout << "can't initialize libsndfile" << std::endl;
        setState ( WAV16_ERROR );
        return false;
    }

    channels = sfh->channels();
    samplerate = sfh->samplerate();

    if ( channels > 2 )
    {
        std::cout << "loop isn't mono/stereo audio" << std::endl;
        setState ( WAV16_ERROR );
        return false;
    }

    sf_count_t frames = sfh->frames ();

    short *dataInArray = new short [ channels * frames ];

    if ( sfh->readf ( dataInArray, frames ) != frames )
    {
        std::cout << "frame mismatch" << std::endl;
        setState ( WAV16_ERROR );
        return false;
    }
    sf_count_t sampleCount = channels * frames;
    for ( sf_count_t i = 0; i < sampleCount; i++ )
    {
        dataIn.push_back ( SampleMath::shortToDouble( dataInArray[i] ));
    }
    delete[] dataInArray;

    if ( dataIn.empty() )
    {
        std::cout << "dataIn empty after fread" << std::endl;
        setState ( WAV16_ERROR );
        return false;
    }
    setState ( WAV16_HAS_DATA_IN );
    return true;
}

void Wav16::generateSine (
    double periods,
    double amplitude,
    size_t frames,
    size_t channelsParameter,
    size_t samplerateParameter )
{
    clear();

    channels = channelsParameter;
    samplerate = samplerateParameter;

    double currentSample, frequency;
    frequency = periods / ( (double) frames * channels );

    for ( size_t i = 0; i < frames; i++ )
    {
        currentSample = 0.6 * DOUBLE_SHRT_MAX * ( ( 1.0 - amplitude ) + amplitude * sin ( 2 * PI * frequency * (double) i ) );
        dataIn.push_back ( currentSample );
        dataIn.push_back ( currentSample );
    }

    setState ( WAV16_HAS_DATA_IN );
}

void Wav16::printInfo()
{
}

//stereo implementation; tbd: mono?
bool Wav16::isZeroCrossing ( int i )
{
    if ( ( i-2 ) < 1 || ( i+3 ) >= ( int ) dataIn.size() )
        return false;

    double a1, a2, b1, b2, c1, c2;
    a1 = dataIn.at ( i-2 );
    a2 = dataIn.at ( i-1 );
    b1 = dataIn.at ( i );
    b2 = dataIn.at ( i+1 );
    c1 = dataIn.at ( i+2 );
    c2 = dataIn.at ( i+3 );

    if ( b1 == 0 && b2 == 0 )
        return true;

    else if ( c1 == 0 && c2 == 0 )
        return false;

    else if ( a1 < 0 && a2 < 0 && b1 > 0 && b2 > 0 )
        return true;

    return false;
}

bool Wav16::write ( char *fname )
{
    if ( getState () != WAV16_HAS_DATA_OUT )
        return false;
    return writeRawData ( fname, dataOut );
}


bool Wav16::writeRawData ( char *fname, const std::vector<double>& data )
{
    if ( getState () != WAV16_HAS_DATA_OUT )
        return false;

    short *outputDataArray = new short [ data.size() ];
    for ( size_t i = 0; i < data.size(); i++ )
    {
        outputDataArray[i] = SampleMath::doubleToShort ( data.at ( i ) );
    }

    SndfileHandle sfhOutfile ( fname, SFM_WRITE, format, channels, samplerate );
    if ( !sfhOutfile )
        return false;

    sfhOutfile.write ( outputDataArray, data.size());

    delete[] outputDataArray;

    return true;
}
